"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const resolve = require("resolve");
const path = require("path");
const reg_suit_util_1 = require("reg-suit-util");
const config_manager_1 = require("./config-manager");
const compare = require("reg-cli");
function isPublisher(pluginHolder) {
    return !!pluginHolder["publisher"];
}
function isKeyGenerator(pluginHolder) {
    return !!pluginHolder["keyGenerator"];
}
function isNotifier(pluginHolder) {
    return !!pluginHolder["notifier"];
}
class RegSuitCore {
    constructor(opt) {
        this._notifiers = [];
        this._pluginHolders = [];
        this.logger = reg_suit_util_1.createLogger();
        this._configManager = config_manager_1.default;
        if (opt && opt.logLevel) {
            this.logger.setLevel(opt.logLevel);
        }
        this.noEmit = !!(opt && opt.noEmit);
    }
    _loadPlugins() {
        if (!this._config.plugins)
            return;
        const pluginNames = Object.keys(this._config.plugins);
        pluginNames.forEach(pluginName => {
            this._loadPlugin(pluginName);
        });
    }
    _loadPlugin(name) {
        let pluginFileName = null;
        try {
            pluginFileName = resolve.sync(name, { basedir: process.cwd() });
        }
        catch (e) {
            this.logger.error(`Failed to load plugin '${name}'`);
            throw e;
        }
        if (pluginFileName) {
            const factory = require(pluginFileName);
            const pluginHolder = factory();
            this._pluginHolders.push(Object.assign({}, pluginHolder, { moduleId: name }));
        }
    }
    createQuestions(opt) {
        const config = this._loadConfig(opt.configFileName);
        const noConfigurablePlugins = [];
        const preparerHolders = [];
        opt.pluginNames.forEach(name => this._loadPlugin(name));
        this._pluginHolders.forEach(h => {
            if (h["preparer"]) {
                preparerHolders.push({ name: h.moduleId, preparer: h["preparer"] });
            }
            else {
                noConfigurablePlugins.push(h.moduleId);
            }
        });
        return [
            ...noConfigurablePlugins.map(pluginName => {
                return {
                    name: pluginName,
                    questions: [],
                    prepare: (inquireResult) => Promise.resolve(true),
                    configured: null,
                };
            }),
            ...preparerHolders.map(holder => {
                const questions = holder.preparer.inquire();
                const boundPrepare = (inquireResult) => holder.preparer.prepare({
                    coreConfig: config.core,
                    logger: this.logger.fork(holder.name),
                    options: inquireResult,
                    noEmit: this.noEmit,
                });
                const configured = (config.plugins && typeof config.plugins[holder.name] === "object") ? config.plugins[holder.name] : null;
                return {
                    name: holder.name,
                    // FIXME
                    // TS4053 Return type of public method from exported class has or is using name 'inquirer.Question' from external module "reg-suit-core/node_modules/@types/inquirer/index" but cannot be named.
                    questions: questions,
                    prepare: boundPrepare,
                    configured,
                };
            }),
        ];
    }
    persistMergedConfig(opt, confirm) {
        const baseConfig = this._loadConfig();
        const mergedConfig = {
            core: opt.core ? Object.assign({}, baseConfig.core, opt.core) : baseConfig.core,
            plugins: Object.assign({}, baseConfig.plugins),
        };
        opt.pluginConfigs.forEach(pc => {
            mergedConfig.plugins[pc.name] = baseConfig.plugins ? Object.assign({}, baseConfig.plugins[pc.name], pc.config) : pc.config;
        });
        this.logger.verbose("Merged configuration: ", mergedConfig);
        if (JSON.stringify(baseConfig) === JSON.stringify(mergedConfig)) {
            // If there is no difference, exit quietly.
            return Promise.resolve();
        }
        if (this.noEmit)
            return Promise.resolve();
        return confirm(mergedConfig).then(result => {
            if (result)
                this._configManager.writeConfig(mergedConfig);
        });
    }
    init(configFileName) {
        this._loadConfig(configFileName);
        this._loadPlugins();
        this._initKeyGenerator();
        this._initPublisher();
        this._initNotifiers();
    }
    _initPlugin(targetPlugin, metadata) {
        const replacedConf = this._configManager.replaceEnvValue(this._loadConfig());
        let pluginSpecifiedOption;
        if (replacedConf.plugins && replacedConf.plugins[metadata.moduleId]) {
            pluginSpecifiedOption = replacedConf.plugins[metadata.moduleId];
        }
        else {
            pluginSpecifiedOption = {};
        }
        targetPlugin.init({
            coreConfig: this._config.core,
            logger: this.logger.fork(metadata.moduleId),
            options: pluginSpecifiedOption,
            noEmit: this.noEmit,
        });
        this.logger.verbose(`${metadata.moduleId} is inialized with: `, pluginSpecifiedOption);
        return targetPlugin;
    }
    _initKeyGenerator() {
        const metadata = this._pluginHolders.filter(holder => isKeyGenerator(holder));
        if (metadata.length > 1) {
            const pluginNames = metadata.map(p => p.moduleId).join(", ");
            this.logger.warn(`2 or more key generator plugins are found. Select one of ${pluginNames}.`);
        }
        else if (metadata.length === 1 && this._config.plugins) {
            const ph = metadata[0];
            if (isKeyGenerator(ph)) {
                this._keyGenerator = this._initPlugin(ph.keyGenerator, ph);
            }
        }
        else {
            this.logger.verbose("No key generator.");
        }
    }
    _initPublisher() {
        const metadata = this._pluginHolders.filter(holder => isPublisher(holder));
        if (metadata.length > 1) {
            const pluginNames = metadata.map(p => p.moduleId).join(", ");
            this.logger.warn(`2 or more publisher plugins are found. Select one of ${pluginNames}.`);
        }
        else if (metadata.length === 1 && this._config.plugins) {
            const ph = metadata[0];
            if (isPublisher(ph)) {
                this._publisher = this._initPlugin(ph.publisher, ph);
            }
        }
        else {
            this.logger.verbose("No publisher.");
        }
    }
    _initNotifiers() {
        const metadata = this._pluginHolders.filter(holder => isNotifier(holder));
        this._notifiers = [];
        metadata.forEach(ph => {
            if (!this._config.plugins)
                return;
            const pluginSpecifiedOption = this._config.plugins[ph.moduleId];
            if (isNotifier(ph)) {
                this._notifiers.push(this._initPlugin(ph.notifier, ph));
            }
        });
    }
    _loadConfig(configFileName) {
        if (!this._config) {
            if (configFileName) {
                this.logger.verbose(`config file: ${configFileName}`);
            }
            else {
                this.logger.verbose(`config file not specified, load from ${config_manager_1.default.defaultConfigFileName}.`);
            }
            this._config = this._configManager.readConfig(configFileName).config;
        }
        return this._config;
    }
    getDirectoryInfo(configFileName) {
        this._loadConfig(configFileName);
        const actualDir = path.join(path.resolve(process.cwd(), this._config.core.workingDir), this._config.core.actualDir);
        const expectedDir = path.join(path.resolve(process.cwd(), this._config.core.workingDir), this._config.core.expectedDir);
        return {
            actualDir,
            expectedDir,
        };
    }
    runAll() {
        return this.getExpectedKey()
            .then(ctx => this.fetch(ctx))
            .then(ctx => this.compare(ctx))
            .then(ctx => this.getActualKey(ctx))
            .then(ctx => this.publish(ctx))
            .then(ctx => this.notify(ctx));
    }
    getExpectedKey() {
        if (this._keyGenerator) {
            return this._keyGenerator.getExpectedKey()
                .then(key => {
                this.logger.info(`Detected the previous snapshot key: '${key}'`);
                return { expectedKey: key };
            })
                .catch(reason => {
                this.logger.warn("Failed to detect the previous snapshot key");
                this.logger.error(reason);
                return Promise.resolve({ expectedKey: null });
            });
        }
        else {
            this.logger.info("Skipped to detect the previous snapshot key because key generator plugin is not set up.");
            return Promise.resolve({ expectedKey: null });
        }
    }
    compare(ctx) {
        return compare({
            actualDir: path.join(this._config.core.workingDir, this._config.core.actualDir),
            expectedDir: path.join(this._config.core.workingDir, this._config.core.expectedDir),
            diffDir: path.join(this._config.core.workingDir, "diff"),
            json: path.join(this._config.core.workingDir, "out.json"),
            report: path.join(this._config.core.workingDir, "index.html"),
            update: false,
            ignoreChange: true,
            urlPrefix: "",
            threshold: .5,
        })
            .then(result => {
            this.logger.verbose("Comparison result:", result);
            return Object.assign({}, ctx, { comparisonResult: result });
        })
            .catch(reason => {
            // re-throw notifiers error because it's fatal.
            this.logger.error("An error occurs during compare images:");
            this.logger.error(reason);
            return Promise.reject(reason);
        });
    }
    getActualKey(ctx) {
        const fallbackFn = () => "snapshot_" + ~~(new Date().getTime() / 1000);
        if (this._keyGenerator) {
            return this._keyGenerator.getActualKey().then(key => {
                if (!key) {
                    this.logger.warn("Failed to generate the current snapshot key.");
                    return Object.assign({}, ctx, { actualKey: fallbackFn() });
                }
                this.logger.info(`The current snapshot key: '${key}'`);
                return Object.assign({}, ctx, { actualKey: key });
            }).catch(reason => {
                this.logger.warn("Failed to gerenate the current snapshot key.");
                this.logger.error(reason);
                return Promise.resolve(Object.assign({}, ctx, { actualKey: fallbackFn() }));
            });
        }
        else {
            const fallbackKey = fallbackFn();
            this.logger.info(`Use '${fallbackKey}' as the current snapshot key because key generator plugin is not set up.`);
            return Promise.resolve(Object.assign({}, ctx, { actualKey: fallbackKey }));
        }
    }
    fetch(ctx) {
        const keyForExpected = ctx.expectedKey;
        if (this._publisher && keyForExpected) {
            return this._publisher.fetch(keyForExpected);
        }
        else if (!keyForExpected) {
            this.logger.info("Skipped to fetch the expeceted data because expected key is null.");
            return Promise.resolve(ctx);
        }
        else if (!this._publisher) {
            this.logger.info("Skipped to fetch the expeceted data because publisher plugin is not set up.");
            return Promise.resolve(ctx);
        }
        else {
            return Promise.resolve(ctx);
        }
    }
    publish(ctx) {
        if (this._publisher) {
            return this._publisher.publish(ctx.actualKey)
                .then(result => {
                this.logger.info(`Published snapshot '${ctx.actualKey}' successfully.`);
                if (result.reportUrl) {
                    this.logger.info(`Report URL: ${result.reportUrl}`);
                }
                this.logger.verbose("Publish result:", result);
                return Object.assign({}, ctx, { reportUrl: result.reportUrl });
            })
                .catch(reason => {
                // re-throw notifiers error because it's fatal.
                this.logger.error("An error occurs during publishing snapshot:");
                this.logger.error(reason);
                return Promise.reject(reason);
            });
        }
        else {
            this.logger.info("Skipped to publish the snapshot data because publisher plugin is not set up.");
            return Promise.resolve(ctx);
        }
    }
    notify(ctx) {
        const notifyParams = Object.assign({}, ctx);
        if (!this._notifiers.length) {
            this.logger.info("Skipped to notify result because notifier plugins are not set up.");
        }
        this.logger.verbose("Notify parameters:", notifyParams);
        return Promise.all(this._notifiers.map((notifier) => {
            return notifier.notify(notifyParams).catch((reason) => {
                // Don't re-throw notifiers error because it's not fatal.
                this.logger.error("An error occurs during notify:");
                this.logger.error(reason);
                return Promise.resolve();
            });
        })).then(() => ctx);
    }
}
exports.RegSuitCore = RegSuitCore;
